package MVC.domain

import MVC.domain.JobsDomain

/**
 * Klasse der Angestelten (employees)
 * 
 * Demonstration von
 *    + before_write
 *    + die Verwendung des internen Loggers
 *    + Die Anwendung von nextvalue (Neuer PK)
 *    + Die Verwendung des Datensatzvalidators (validate)
 *    + Feldvalidator nullable
 * 
 */
class EmployeesDomain extends framework.database.DAO.DataAccessObject {


  /**
   * Das sind die Felder, welche der Applikation bekannt sind
   * 
   * Achtung nicht alle Felder muessen eine Entsprechung in der Datenbank
   * haben. 
   * 
   * Gueltige Datentypen fuer Datenbankrelevante Felder
   *  + Integer
   *  + Long
   *  + Date
   *  + Double
   *  + String
   * 
   */

   Integer  employeeID
   String   firstName
   String   lastName
   String   email
   String   phoneNumber
   Date     hireDate
   String   jobID
   Double   salary
   Integer  managerID
   Integer  departmentID


  /**
   * Die Datenstruktur meta beschreibt die Verbindung 
   * zur Datenbank und das Standardverhalten.
   */
   def meta = [
       /**
        * Beschreibt die Eindeutige Kennung (meist Primarykey) 
        * des Datensatzes.
        * Es wird der Feldname in der Datenbank ('db') und der 
        * Attributname in der Klasse ('domain') deklariert.
        *
        */
       idfield:[db:'EMPLOYEE_ID',domain:'employeeID'],

      /**
       * Tabellenname in der Datenbank
       */
       tablename: 'EMPLOYEES',

       /**
        * Diese Routine (Closure) wird vor dem Schreiben eines 
        * Datensatzes aufgerufen.
        *
        */
       before_write:{param->
          /**
           * dem Closure wird vom Framework
           * eine Map uebergeben. Dzg. beinhaltet diese
           * nur den Wert 'mode' dieser zeigt die Datenbankoperation 
           * (insert, delete und update) an.
           * Das DAO Objekt besitzt entsprechende Konstanten. Es wird 
           * dringend empfohlen diese zu verwenden!
           */
          switch (param.mode) {
             case MODE_INSERT :
                this.employeeID = this.nextval()
                this.log.info "\n*\n* Neue ID#: ${this.employeeID} \n*"
                break
             case MODE_DELETE :
                this.log.info "\n\n*\n* Datensatz mit ID#${this.employeeID} ${this.lastName} wurde geloescht\n*\n" 
                break
             case MODE_UPDATE :
                this.log.info "\n\n*\n* Datensatz mit ID#${this.employeeID} ${this.lastName} wurde veraendert\n*\n" 
                break
             default :
                break
             }
          this.log.info "Demo before_write dboperaton: ${param.mode} (employees domain): '${this.employeeID}'"
          },
       /**
        * Demonstrier die Verwendung das
        * automatische setzen des Primariy-Keys
        * mit nchsthheren Wert.
        * HINT:
        *   Normalerweise wird man hier in Oracle einen Sequenzer
        *   oder bei mySQL den Autokey verwenden.
        */
       nextval:{
          Integer id = (db.firstRow('select MAX(EMPLOYEE_ID)+1 from EMPLOYEES')[0]).toInteger()
          if (id == null) id = 1        
          return id
          },

       /**
        * Ueberprfungen durchfhren
        * Niemand darf unter $1.000,- verdienen
        * 
        * Diese Closure wird nach der Uebernahme der Daten
        * aus dem CGI im Datensatz ueberprft.
        * 
        * Sie wird fuer Pruefungen auf Datensatzebene verwendet.
        * 
        */
       validate : {
          if ((this.salary?:0.0) < 1000.0) return "Gehalt muss gr&ouml;sser oder gleich \$1.000,- sein!"
          return ''
          },

      /**
       * Verbindung zwischen Datenbank und 
       * Domain herstellen.
       * Zu den Feldern werden noch Type und
       * Moeglicherweise Verhalten definiert werden.
       * 
       * HINT
       *    In der Datenstruktur fiels werden die Inahlte
       *    als String uebergeben. Damit muss eine Umwandlung
       *    in den gewuenschten Datentyp erfolgen.
       *    ACHTUNG: Beachte, dass es Datentypenfehler durch
       *    fehlerhafte Benutzereingaben geben kann.
       *
       */
      fields: [
         employeeID:   [
            dbfield:'EMPLOYEE_ID',
            type:'Integer',
            ],
         firstName:     [
            dbfield:'FIRST_NAME',
            nullable:false, /* Feld muss Inhalt haben (not NULL) */
            validate:{value->
               if (value.size() < 4) return 'Der Name muss mindestens 4 Buchstaben haben'
               return ''
               },
            type:'String'
            ],
         lastName:      [
            dbfield:'LAST_NAME',
            nullable:false, /* Feld muss Inhalt haben (not NULL) */
            type:'String'
            ],
         email: [
            dbfield:'EMAIL',
            type:'String',
            nullable:false /* Feld muss Inhalt haben (not NULL) */
            ],
         phoneNumber:   [  /* Dieses Feld darf NULL sein */
            dbfield:'PHONE_NUMBER',
            type:'String'
            ],
         hireDate :     [  /* Dieses Feld darf NULL sein */
            dbfield:'HIRE_DATE',
            type:'Date'
            ],
        /** 
         * Feldmethode get:
         * Vor dem schreibe wird der Input auf Dezimalkomma geprueft 
         * und ggf. durch Punkt ersetzt 
         * 
         * Feldmethode set:
         * Nach dem lesen wird der Inhalt gendert 
         */
         salary:[
            dbfield:'SALARY',
            type:'Double',

           /**
            * Diese Closure wird nach dem Lesen der Daten aus dem
            * Datenbakrecord aufgerufen. Der Returnwert wird in 
            * das Feld der Domain geschrieben.
            */
            after_read:{value->return value?:0.0},

           /**
            * Diese Closure wird vor dem Schreiben der Daten
            * in den Datenbakrecord aufgerufen und das Ergebnis
            * in das Datenfeld geschrieben.
            */
            before_write:{value->return value * 1.0},

           /**
            * Diese Closure wird vor dem 
            * schreiben aufgerufen.
            */
            after_cgi:{value->return (value =~ /,/).replaceFirst('.')}
            ],

         jobID:         [
            dbfield:'JOB_ID',
            type:'String'
            ],

         managerID:     [
            dbfield:'MANAGER_ID',
            type:'Integer'
            ],

         departmentID:  [
            dbfield:'DEPARTMENT_ID',
            type:'Integer'
            ]
         ]
      ]

     /**
      * Liefert eine Datasource
      * fuer Selectbox 
      * 
      * @param employeeID     zu selectierender Satz
      * 
      * @return List
      *
      */
      List selectDatasource(Integer employeeID) {
         List datasource = []
         this.eachDomain('(1=1)','LAST_NAME,FIRST_NAME') {dom->
            datasource.add([
               dom.employeeID,
               "${dom.lastName} ${dom.firstName} (${dom.employeeID})",
               (dom.employeeID == employeeID)
               ])
            }
         return datasource
         }
   }