/** Datenbanuzgriffsobjekt 
 *
 * <tt></pre>
 *
 * -- wann --+-wer-+-- was ----------------------------------------------
 * 01.11.2007| ny  | estellt
 * 02.11.2007| ny  | Sourcedocu
 * 09.11.2007| ny  | idfield auf Datenbankfeld erweitert (idfield:[db:...])
 * 15.11.2007| ny  | setFromCgi mit "" = NULL versehen (EMPTY_IS_NULL)
 * 16.11.2007| ny  | each eingebaut
 * 30.11.2007| ny  | Erweiterung getfromcgi: validiereung
 * 18.12.2007| ny  | Datumsbehandlung Oracle
 * 18.01.2008| ny  | Closure validate wird nicht mehr 
 *           |     | untersttzt
 * 31.01.2008| ny  | Include in Save eingebaut 
 * 05.02.2008| ny  | Include in setFromCgi eingebaut 
 * 12.02.2008| ny  | eachDomain verbessertes
 *           |     | BO Handling
 * 25.03.2008| ny  | einbau der CGI Option bei Domain
 *           |     | um z.B. Checkbox handeln zu knnen
 * 08.07.2008| ny  | SQL.Logger eingebaut
 * 13.09.2008| ny  | Methode allFields eingebaut 
 * 30.09.2008| ny  | eachDomainWithLimit eingebaut 
 * 28.10.2008| ny  | after_write Event eingebaut
 * 31.10.2008| ny  | Refaktoring: Einbau von setValueInBo
 * 01.11.2008| ny  | Commit/Rollback 
 * 16.11.2008| ny  | Dynamische Abfragen bei eachDomain eingebau
 * 22.11.2008| ny  | Erweiterung um QueryBuilder fuer eachDomain
 *           |     | Parameter criteria.
 *           |     | 
 *           |     | 
 *
 * Autor: W. Nagy
 *
 * Hintergrund<br/>
 *
 * Dieses Modul implementiert Data Access Objects (DAO). DAOs Kapseln den
 * Datenbankzugriff.
 *
 * Die Klasse ist abstrakt deklariert und kann somit nicht direkt instanziert werden
 *
 * Arichitektur:
 * 
 * .------------------------------------------.
 * | APPLIKATION:                             |
 * |   kennt nur Domains                      |
 * |                                          |
 * .------------------------------------------.
 * | PERSISTENZSCHICHT(DAO)                   |
 * |  Kennt Tabellen- und Feldnamen           |
 * |                                          |
 * '------------------------------------------'
 * | LOWLEVEL (Klasse Database)               |
 * |  Kennt Datebankkverbindungen und         |
 * |  Kapselt Lowlevel Zugriffe auf GSQL      |
 * '------------------------------------------'
 * | GSQL  (Db-ZUgriffe auf GSQL)             |
 * |                                          |
 * |                                          |
 * '------------------------------------------'
 * | JDBC                                     |
 * |                                          |
 * |                                          |
 * '------------------------------------------'
 *
 *
 * </abstract>
 * .- globals -.
 * --- Name ------------+--- Beschreibung -------------------+-- Vorgabe ------
 * db                   | Datenbankhandle                    | null
 *                      | Ist eine Instanz von               |
 *                      | groovy.sql.Sql                     |
 *                      |                                    |
 * lastSql              | Speichert die letzte SQL Anweisung | null
 *                      |                                    |
 * writeMode            | Die Methode save kann selbst-      | null
 *                      | staendig die Schreibende Methode   |
 *                      | ermitteln (update/insert/ usw.     |
 *                      |                                    |
 * Metadaten
 * ---------
 *
 * Die Klasse muss informationen ueber die Datenbanktabelle haben.
 *
 * Diese Informatinen werden in einer Datenstruktur Namens meta gespeichert.
 * Beispiel:
 *
 * class Jobs extends DataAccessObject {
 *   String  jobID
 *   String  jobTitle
 *   Double  minSalary
 *   Double  maxSalary
 *
 *   def meta = [
 *              idfield:[db:'JOB_ID'],
 *              fields:[
 *                     jobID:[dbfield:'JOB_ID'],
 *                     jobTitle:[dbfield:'JOB_TITLE'],
 *                     minSalary:[dbfield:'MIN_SALARY'],
 *                     maxSalary:[dbfield:'MAX_SALARY']
 *                     ]
 *              ]
 *   }
 *
 *
 *
 * + Tabellenname:
 *   in den Metadaten muss das Property tablename unbedingt angegeben werden.
 *
 * + idfield:
 *    - db
 *       Die Datensaetze in der Datenban muessen eine einedeutige Kennung haben.
 *       Die Kennung muss in einem einzelnen Feld stehen.
 *       Der Name des Feldes wird im der Option idfield vermerkt werden.
 *    -domain
 *       Feldname des Kennzeichnungsfeldes in der Domain (dzg nicht in verwendung)
 *
 * + fields:
 *   Die persistenten Felder werden in der Liste fields notiert.
 *   Alle Felder der Klasse welche nicht in der Feldliste stehen werden
 *   transistent behandelt.
 *
 *   Die Einfachste Form der Felddeklaration ist feldname:null. Die Klasse nimmt an,
 *   dass das Feld gleich wie inder Datenbanktabelle heisst.
 *
 *   Beispiel:
 *
 *    ... fields:[
 *
 *                      FIRST_NAME    :null,
 *                      LAST_NAME     :null,
 *
 *   - Fieldsparameter
 *     dbfield: Name des Datenbankfeldes
 *     pk:      [true|false] Feld ist ein Primariy key (for later use)
 *     HINT: diese Liste wird noch erweitert
 *
 * Glossar
 * -------
 * --- Abkuerzung ----------+--- meint -------------------------
 * BO                       | Buisness Objekt (auch Domain genannt)
 * POGO                     | Plain old Groovy Objekt
 *                          | Eine Klasse mit den Felder des BOs
 *                          | und den Metadaten (meta)
 * ID                       | Eindeutige Kennung des BOs
 *                          |
 *                          |
 *                          |
 *
 * helper
 * ------
 *
 * die Klasse DbHelper stellt die physische Verbindung zur Datenbank her
 * Mit hilfe der klasse knnen Datebnankverbindungen mehrfach verwendet werden.
 *
 *      dbh = new DbHelper().connect(url:'jdbc:hsqldb:file:c:/temp/employee',
 *                                   username:'sa',
 *                                   password:'',
 *                                   driverClassName:'org.hsqldb.jdbcDriver'
 *                                  )
 *
 *      employees = new Employees(db:dbh)
 *      jobs = new Jobs(db:dbh)
 *
 *
 *
 *
 * <tt></pre>
 *
 * -- wann --+-wer-+-- was ----------------------------------------------
 * 01.11.2007| ny  | estellt
 * 02.11.2007| ny  | Sourcedocu
 * 09.11.2007| ny  | idfield auf Datenbankfeld erweitert (idfield:[db:...])
 * 15.11.2007| ny  | setFromCgi mit "" = NULL versehen (EMPTY_IS_NULL)
 * 16.11.2007| ny  | each eingebaut
 * 30.11.2007| ny  | Erweiterung getfromcgi: validiereung
 * 18.12.2007| ny  | Datumsbehandlung Oracle
 * 18.01.2008| ny  | Closure validate wird nicht mehr 
 *           |     | untersttzt
 * 31.01.2008| ny  | Include in Save eingebaut 
 * 05.02.2008| ny  | Include in setFromCgi eingebaut 
 * 12.02.2008| ny  | eachDomain verbessertes
 *           |     | BO Handling
 * 25.03.2008| ny  | einbau der CGI Option bei Domain
 *           |     | um z.B. Checkbox handeln zu knnen
 * 08.07.2008| ny  | SQL.Logger eingebaut
 * 13.09.2008| ny  | Methode allFields eingebaut 
 * 30.09.2008| ny  | eachDomainWithLimit eingebaut 
 * 28.10.2008| ny  | after_write Event eingebaut
 * 31.10.2008| ny  | Refaktoring: Einbau von setValueInBo
 * 01.11.2008| ny  | Commit/Rollback 
 * 16.11.2008| ny  | Dynamische Abfragen bei eachDomain eingebau
 *           |     | 
 *           |     | 
 *
 * Autor: W. Nagy
 *
 * Hintergrund<br/>
 *
 * Dieses Modul implementiert Data Access Objects (DAO). DAOs Kapseln den
 * Datenbankzugriff.
 *
 * Die Klasse ist abstrakt deklariert und kann somit nicht direkt instanziert werden
 *
 * Arichitektur:
 * 
 * .------------------------------------------.
 * | APPLIKATION:                             |
 * |   kennt nur Domains                      |
 * |                                          |
 * .------------------------------------------.
 * | PERSISTENZSCHICHT(DAO)                   |
 * |  Kennt Tabellen- und Feldnamen           |
 * |                                          |
 * '------------------------------------------'
 * | LOWLEVEL (Klasse Database)               |
 * |  Kennt Datebankkverbindungen und         |
 * |  Kapselt Lowlevel Zugriffe auf GSQL      |
 * '------------------------------------------'
 * | GSQL  (Db-ZUgriffe auf GSQL)             |
 * |                                          |
 * |                                          |
 * '------------------------------------------'
 * | JDBC                                     |
 * |                                          |
 * |                                          |
 * '------------------------------------------'
 *
 *
 * </abstract>
 * .- globals -.
 * --- Name ------------+--- Beschreibung -------------------+-- Vorgabe ------
 * db                   | Datenbankhandle                    | null
 *                      | Ist eine Instanz von               |
 *                      | groovy.sql.Sql                     |
 *                      |                                    |
 * lastSql              | Speichert die letzte SQL Anweisung | null
 *                      |                                    |
 * writeMode            | Die Methode save kann selbst-      | null
 *                      | staendig die Schreibende Methode   |
 *                      | ermitteln (update/insert/ usw.     |
 *                      |                                    |
 * Metadaten
 * ---------
 *
 * Die Klasse muss informationen ueber die Datenbanktabelle haben.
 *
 * Diese Informatinen werden in einer Datenstruktur Namens meta gespeichert.
 * Beispiel:
 *
 * class Jobs extends DataAccessObject {
 *   String  jobID
 *   String  jobTitle
 *   Double  minSalary
 *   Double  maxSalary
 *
 *   def meta = [
 *              idfield:[db:'JOB_ID'],
 *              fields:[
 *                     jobID:[dbfield:'JOB_ID'],
 *                     jobTitle:[dbfield:'JOB_TITLE'],
 *                     minSalary:[dbfield:'MIN_SALARY'],
 *                     maxSalary:[dbfield:'MAX_SALARY']
 *                     ]
 *              ]
 *   }
 *
 *
 *
 * + Tabellenname:
 *   in den Metadaten muss das Property tablename unbedingt angegeben werden.
 *
 * + idfield:
 *    - db
 *       Die Datensaetze in der Datenban muessen eine einedeutige Kennung haben.
 *       Die Kennung muss in einem einzelnen Feld stehen.
 *       Der Name des Feldes wird im der Option idfield vermerkt werden.
 *    -domain
 *       Feldname des Kennzeichnungsfeldes in der Domain (dzg nicht in verwendung)
 *
 * + fields:
 *   Die persistenten Felder werden in der Liste fields notiert.
 *   Alle Felder der Klasse welche nicht in der Feldliste stehen werden
 *   transistent behandelt.
 *
 *   Die Einfachste Form der Felddeklaration ist feldname:null. Die Klasse nimmt an,
 *   dass das Feld gleich wie inder Datenbanktabelle heisst.
 *
 *   Beispiel:
 *
 *    ... fields:[
 *
 *                      FIRST_NAME    :null,
 *                      LAST_NAME     :null,
 *
 *   - Fieldsparameter
 *     dbfield: Name des Datenbankfeldes
 *     pk:      [true|false] Feld ist ein Primariy key (for later use)
 *     HINT: diese Liste wird noch erweitert
 *
 * Glossar
 * -------
 * --- Abkuerzung ----------+--- meint -------------------------
 * BO                       | Buisness Objekt (auch Domain genannt)
 * POGO                     | Plain old Groovy Objekt
 *                          | Eine Klasse mit den Felder des BOs
 *                          | und den Metadaten (meta)
 * ID                       | Eindeutige Kennung des BOs
 *                          |
 *                          |
 *                          |
 *
 * helper
 * ------
 *
 * die Klasse DbHelper stellt die physische Verbindung zur Datenbank her
 * Mit hilfe der klasse knnen Datebnankverbindungen mehrfach verwendet werden.
 *
 *      dbh = new DbHelper().connect(url:'jdbc:hsqldb:file:c:/temp/employee',
 *                                   username:'sa',
 *                                   password:'',
 *                                   driverClassName:'org.hsqldb.jdbcDriver
 *                                  )
 *
 *      employees = new Employees(db:dbh)
 *      jobs = new Jobs(db:dbh)
 *
 * Dynamische Abfragen
 * -------------------
 * 
 * Um eine Trenneung der Datenbankelder von den Domainfeldern zu 
 * ermoegliche, wurde die Methode eachDomain implementiert.
 * 
 * Syntax
 * ------
 * Allgemeine Form:
 * eachDomainBy<Domain-Fieldname&<Condition[<Log-Operator>],...]
 *
 * Feldname
 * 
 * Um moegliche Namenskonflikte zu vermeinden koennen Feldnamen mit grossem Anfangsbuchstaben
 * notiert werden. Beispiel: "JobID" entspricht "jobID".
 * 
 * Condition
 * 
 *   + Equal
 *   + NotEqual
 *   + LessThan
 *   + LessThanEquals
 *   + GreaterThan
 *   + GreaterThanEquals
 *   + Like
 *   + NotLike
 *   + IsNull
 *   + IsNotNull
 * 
 * Wenn keine Condition angegeben wurde, wird vom System Equal angenommen (Siehe Beispiel 2).
 *
 * Log-Operatoren
 * --------------
 * Als logische Operatoren sind And und Or zugelassen.
 * Klammerung ist nicht m&ouml;glich.
 * 
 * Moegliche Komplikationen
 * 
 * Namenskonflikte:
 *
 * Das System versucht die Feldnamen und die Operationen aus dem String zu "eraten". Dabei ist die Gross und Kleinschreibung
 * "Camelstyle" behilflich. Es besteht die aber die Moeglichkeit, dass in Feldnamen aehnliche Muster enhalten sind. 
 * "salaryLessThan" "minAndere" etc. Dann gibt es 'Broeseln'.
 *
 * Werte:
 *
 * Die Abfragen werden in prepared statments formuliert und der Inhalt der conditions Option uebergeben. Dies kann zu Schwierigkeiten 
 * fuehren, da der konkrete JDBC Treiber nicht immer eine Korrekte Typenkonvertion durchfuehrt.
 * Der Entwickler muss moeglicherweise selbst fuer die geeignete Umwandlung sorgen.
 *
 */
package framework.database.DAO

import groovy.sql.Sql
import java.util.logging.*

import framework.helper.DateConverter
import framework.database.DAO.helper.QueryBuilder
import framework.database.DAO.helper.*

import framework.helper.Logging

// =====================================================================
 abstract class DataAccessObject {
// =====================================================================

  /**
   * Definition Konstanten
   */
   public final static String MODE_INSERT  = 'insert'
   public final static String MODE_UPDATE  = 'update'
   public final static String MODE_DELETE  = 'delete'
   public final static String MODE_READ    = 'read'
   public final static String MODE_UNKNOWN = 'unknown'
   public final static String MODE_ERROR   = 'error'

   public final static FORM_ACEPT_CHARSET_UTF_8 = 'UTF-8'
   public final static FORM_ACEPT_CHARSET_ISO_8859_1 = 'ISO-8859-1'

   String formAceptCharset = FORM_ACEPT_CHARSET_UTF_8


  /**
   * Attribute
   */
   Sql  db                       // Datenbankhandle
   String lastSql = ''           // buffer fuer das letzte ausgefuehrte SQL Statment
   String writeMode = ''         // writeMode ::= [unkonwn|read|insert|update|delete]
   String sqlError = ''          // SQL Error

   boolean EMPTY_IS_NULL = true  // Definiert, dass "Leerfelder"  welche ueber das
                                 // CGI geliefert werden in der Datenbank als
                                 // NULL gespeichert werden.
                                 // Dies Variable wird in der Methode setFromCgi verwendet.

   boolean DECIMALPOINT_IS_COMMA = true // Bei Double wird automatisch eine 
                                // Konvertierung von Komma mit Dezimalpunkt vorgenommen

   final static String VERSION = '2.01E040'

   List errors = []              // Wird vom setFromCgi verwendet und listet Fehlermeldungen

   def log = null               // Logger Assoziation
   Boolean writeLogMessage = true   // um temporaeren ein und ausschalten des Loggings
   Boolean isOk = true

   Logger logger = Logger.getLogger('DateAccessObojekct')

   // --- usefull onliner ----------------------------------------------
   String getTablename()         {this.meta?.tablename}
   void open(param)              {this.db = new DbHelper().connect(param)}
   def firstRow(String sql)      {db.firstRow(sql)}
   String getDbfieldname(field)  {field.value?.dbfield == null ? field.key : field.value?.dbfield}
   def connect(param)            {return new DbHelper(param)}
   void beginTransaction()       {this.db.useConnection.autoCommit=false}
   void commit()                 {this.db.commit()}
   void rollback()               {this.db.rollback()}
   void autocommit()             {this.db.useConnection.autoCommit=true}
   void clearErrors()            {this.errors = [];this.isOk=true}
   void isOK()                   {this.isOk = true}
   void isNotOK()                {this.isOk = false}
   String idFieldName()          {this.meta?.idfield['db']}
   Boolean hasErrors()           {return !this.isOk}
   // ------------------------------------------------------------------


  /**
   * Benhandlung von dynamischen Methoden.
   * 
   * Dynamische Methoden sind Methodenaufrufe, welche Parameterangaben
   * im Namen haben. z.B: eachDomainByJobID(100).
   * 
   */
   def methodMissing(String name, args) {
      
      if (!(name =~ /^eachDomainBy/)) {
         throw new Exception("${this.getClass().getName()}::kann die Methode ${name} nicht finden.")
         }
      doEachDomainBy(name,args)
    }


  /**
   * Sucht Db-Feldnamen fuer Dynamische Abragen
   *
   * @param    fieldname
   *
   * @return   columnname 
   * 
   * Kann der feldname nicht aufgelst werden, wird
   * null zurueckgeliefert
   * 
   */
   String getColumnnameFromDomainfield(String field) {

      String columnname = null

      if (this.meta?.fields.get(field) != null) {
         columnname = this.meta?.fields[field]?.get('dbfield')
         }
      else {
         switch (field) {
            case null : break
            case {field.size() == 1} :
               field = field.toLowerCase()
               break
            default  :
               field = field[0].toLowerCase()+field[1..field.size()-1]
               break
            }
         columnname = this.meta?.fields[field]?.get('dbfield')
         }
      if (columnname == null)
         throw new Exception("In der dynamischen Anweisung kann Feld '${field}' nicht zu DB-Feld aufgeloest werden.")

      return columnname
      }

  /**
   * Bearbeiten der dynamische Methode
   * eachDomainBy...
   * 
   * Folgende Operatoren stehen zur Verfuegung
   * 
   * - Equal
   * - NotEqual
   * - GreaterThan
   * - GreaterThanEquals
   * - LessThan
   * - LessThanEquals
   * - Like
   * - NotLike
   * - IsNull
   * - IsNotNull
   * 
   * Der Voragabewert ist Equal
   * 
   * 
   */
   void doEachDomainBy(String name, args) {
      name = (name =~ /^eachDomainBy/).replaceFirst('')
       
      List clauses = []
      List fields = name.split(/And|Or/)
 
      name.eachMatch(/And|Or/) {clauses<<it[0]}
      
      String  query = ''
      Integer iClauses = 0
      Integer iOperator = 0
      List    values = []

      fields.each{field->
         String logOperator = iOperator >= clauses.size() ? '' : clauses[iOperator]?:''.toLowerCase()
         iOperator++
         List conditions = [] 
         String domainField = ''
         if (iClauses < args[0].conditions.size())
            values<<args[0].conditions[iClauses]

         switch (field ) {
            case ~/.*Equal$/ :
               domainField = (field =~/Equal$/).replaceFirst('')
               query += "${getColumnnameFromDomainfield(domainField)} = ? ${logOperator} "
               iClauses++
               break

            case ~/.*NotEqual$/ :
               domainField = (field =~/Equal$/).replaceFirst('')
               query += "${getColumnnameFromDomainfield(domainField)} <> ? ${logOperator} "
               iClauses++
               break
 
            case ~/.*GreaterThan$/ :
               domainField = (field =~/GreaterThan$/).replaceFirst('')
               query += "${getColumnnameFromDomainfield(domainField)} > ? ${logOperator} "
               iClauses++
               break
 
            case ~/.*GreaterThanEquals$/ :
               domainField = (field =~/GreaterThanEquals$/).replaceFirst('')
               query += "${getColumnnameFromDomainfield(domainField)} >= ? ${logOperator} "
               iClauses++
               break

            case ~/.*LessThanEquals$/ :
               domainField = (field =~/LessThanEquals$/).replaceFirst('')
               query += "${getColumnnameFromDomainfield(domainField)} <= ? ${logOperator} "
               iClauses++
               break


            case ~/.*LessThan$/ :
               domainField = (field =~/LessThan$/).replaceFirst('')
               query += "${getColumnnameFromDomainfield(domainField)} < ? ${logOperator} "
               iClauses++
               break

            case  ~/.*NotLike$/ :
               domainField = (field =~/NotLike$/).replaceFirst('')
               query += "${getColumnnameFromDomainfield(domainField)} NOT LIKE ? ${logOperator} "
               iClauses++
               break            

            case  ~/.*Like$/ :
               domainField = (field =~/Like$/).replaceFirst('')
               query += "${getColumnnameFromDomainfield(domainField)} LIKE ? ${logOperator} "
               iClauses++
               break            

            case ~/.*IsNotNull$/ :
               domainField = (field =~/IsNotNull$/).replaceFirst('')
               query += "${getColumnnameFromDomainfield(domainField)} IS NOT NULL ${logOperator} "
               break

            case ~/.*IsNull$/ :
               domainField = (field =~/IsNull$/).replaceFirst('')
               query += "${getColumnnameFromDomainfield(domainField)} IS NULL ${logOperator} "
               break
  
            default :
               query += "${getColumnnameFromDomainfield(field)} = ? ${logOperator} "
               iClauses++
               break
            }
         }

      List orderby = []
      args[0].orderby.each{
         orderby<<getColumnnameFromDomainfield(it)
         }
      
      String sql = "select * from ${this.meta?.tablename} where ${query} ${orderby.size() > 0 ? 'order by '+orderby.join(',') : ''}"

      // Datenstrom durchlaufen
      //
      Integer offset    = args[0].get('offset',0)
      Integer max       = args[0].get('max',null)
      Long cntRec = max?:0L
      Long cntOff = offset?:0L

      this.lastSql = "${sql} ${values}"
      db.eachRow(sql,values) {row->
        /**
         * Ueberlesen wenn Offset gewnscht
         */
         if (cntOff > 0) {
            cntOff--
            return
            }
         
        /**
         * Wenn maximale Anzal gewuenscht
         * wird Recordzaehler minimiert
         * Wenn nicht Zhler minimieren
         */
         if(max == null) 
            cntRec = 1
         else 
            cntRec--
         
         if (cntRec >= 0) {
            fillDomainFields(row)
            this.meta?.after_read?.call()
            // Aufrufen Closure
            args[args.size()-1].call(this)
            }
         }
      }

  /**
   *  
   *  Schreibt Logeintrag
   *  
   *  HINT:
   *     Zum Schreiben eines Logeintrags muss dem 
   *     Objekt ein Logger mitgeben werden. Dieser wird beim
   *     Initialisieren des Objekts angelegt.
   *     Beispiel:
   *     
   *        def auftrag = new AuftragDomain(
   *                         db:db.dbh,
   *                         log:new AuftragLogging(request:request)
   *                         )
   */     
   void writeLog (String message) {
      if (log != null && writeLogMessage) {
         log.info(message)
         }
      }

  /**
   * 
   *  Prueft ob das uebergebne Feld NULL ist
   * 
   *  @param:  value          Feldinhalt
   * 
   *  @return: true/false     Liefert Wahrheitswert wenn das Feld logisch NULL ist
   * 
   */
   Boolean isNull (value ) {

      if (value == null) return true

      if (value =~ /^\s*$/) {
         return true}
      else {
         return false
         }
      }

  /**
   * 
   *  Wenn Value logisch NULL ist so
   *  wird die Fehlermeldung ausgegeben und ein Wahrheitswert
   *  Zurckgegeben
   * 
   *  @param:  value          Wert
   *  @param:  field          Name des Feldes
   * 
   *  @return: Wahrheitswert
   */ 
   def notNull (value,field) {

      if (isNull(value)) {  // Leerfeld
         return false
         }
      else {
         return true
         }
      }

  /**
   * 
   *  Liefer bei Oracledatebank Sequenzer
   * 
   *  @param: (optional) name des Sequencers
   * 
   */
   Object nextval(name) {
      if (name == null) {  //Sequencername wurde nicht uebergeben
         name = this.meta.sequencer
         if (name == null) {  // Kein Sequencername definiert
            if (this.meta.nextval != null) { // aber Nextval Clousure
               // Durchfhren Clousure
               return this.meta.nextval.call()
               }
            }
         }
      
      String sql = "select ${name}.nextval from dual"
      return db.firstRow(sql)[0]
      }

  /**
   * 
   *  Shortcut zu eachRow
   * 
   *  usage: eachRow('select ...') {row->print row.LAST_NAME}
   * 
   *  @param:  sql         Selectstatment
   *  @param:  clousure    Methoden welche auf die Rows angewendet werden
   * 
   */ 
   void eachRow(String sql,Closure closure) {
      db.eachRow(sql) {
         closure(it)
         }
      }

  /**
   * 
   *  Ist ein Shortcut auf die Nativ
   * 
   */
   def firstRow = {param->
      return db.firstRow(param)
      }

  /**
   * 
   * Durchlaeuft die Metadaten der Domain
   * und gibt eine Liste von Feldnamen zurck (SQL Listclause)
   * 
   * Date Datentyp OraDate erfaehrt eine besondere Behandlung
   * Da Oracle nur Date Datum ohne den Zeitanteil liefert wird
   * ein String im Format CCYY-MM-DD HH:MI zurckgeliefert.
   * 
   */ 
   String giveFieldList () {
      List fieldlist = []

      this.meta?.fields.each {field->
         String fieldname = field.value?.dbfield == null ? field.key : field.value?.dbfield
         String fieldtype = field.value?.type == null ? 'String' : field.value?.type
         switch (fieldtype) {
            case 'OraDate' :
               fieldlist.add("to_char(${fieldname},'YYYY-MM-DD HH24:MI') as \"${fieldname}\"")
               break
            default :
               fieldlist.add(fieldname)
            }
         }
      return fieldlist.join(',')
      }

  /**
   * Befuellen Buisiness Object mit einem 
   * Datenbankrecord
   * 
   * @param    rec      Record 
   * 
   * HINT:  
   *     Es werden keine Validatoren aufgerufen!
   *     Es werden Typenkonvertierungen (speziell bei DATE)
   *     von der Datenbank zur Domain durchgefuehrt.
   *     
   */
   void fillDomainFields(rec) {
            
      this.meta?.fields.each {field->
         String fieldname = field.value?.dbfield == null ? field.key : field.value?.dbfield

        /**
         * Zentrales Wertfeld
         */
         def fieldvalue

         // Ist Rec Resultset
         if (rec instanceof groovy.sql.GroovyResultSet) {
            Integer colInx = rec.findColumn(fieldname).toInteger()-1
            // befuellen Datenfelds
            fieldvalue = rec[colInx]
            }
         else {
            fieldvalue = rec.get(fieldname)
            }

        /**
         * In der Domain kann fuer jedes Feld kann eine
         * Routine definiert werden, welche nach dem Lesen
         * aus der Datenbak aufgerufen wird ("after_read").
         * 
         * In dieser knnen ndern Aenderungen vorgenommen werden.
         * 
         * usage:
         *  Liefert nach dem lesen 0.0 wenn NULL in der 
         *  Datenbank steht:
         *
         *  after_read:{value-> return value?:0.0}
         *  
         */
         if (field.value?.after_read != null) {
            try {
               fieldvalue = field.value.after_read.call(fieldvalue)
               }
            catch(Exception ex) {
               throw new Exception("In Domain: ${this.getClass().getName()} im after_read des Feldes ${fieldname} mit Wert '${fieldvalue}' trat ein Fehler auf ${ex.message}")
               }
            }

        /**
         * Es wird der Inhalt des Datenbankerecors in
         * das Buissnes Objekt kopiert.
         */
         if (fieldvalue != null){

            try {
               // Umwandlung von Datenbankdatentypen
               // in Domainfelder
               //

               switch (field.value?.type) {
                  // Bearbeiten der Unterschiedlichen Datums und Zeitdarstellunge
                  // in verschiedener Datenbanken
                  //
                  case 'OraDate' :
                     switch (fieldvalue) {
                        case java.sql.Date :
                           try{
                              this.setProperty(field.key,new java.sql.Timestamp(fieldvalue.getTime()))
                              }
                           catch (Exception e) {
                              throw new Exception ('(1) DAO: [OraDate] Timestampumwandlungsproblem (java.sql.Date): '+e.message+"\n")
                              return false
                              }
                           break
                        default:  
                           try {
                              this.setProperty(field.key, new DbHelper().OraDateStringToDate(fieldvalue))
                              }
                           catch(Exception ex) {
                              throw new Exception('(2) DAO: [OraDate] Timestampumwandlungsproblem (java.sql.Date): '+ex.message+"\n")
                              return false
                              }
                           break
                        }
                     break
                  case 'Date' :
                     // Es wird eine SQL Konforme Datumsdarstellungen angliefert
                     //
                     switch (fieldvalue) {
                        // 'Standard SQL Daten
                        // HINT: in Oracle Typ: DATE
                        //
                        case java.sql.Date :
                           try{
                              this.setProperty(field.key,new java.sql.Timestamp(fieldvalue.getTime()))
                              }
                           catch (Exception e) {
                              throw new Exception ('(1) DAO: Timestampumwandlungsproblem (java.sql.Date): '+e.message+"\n")
                              return false
                              }
                           break
                        case java.sql.Timestamp :
                           try{
                              this.setProperty(field.key,new java.sql.Timestamp(fieldvalue.getTime()))
                              }
                           catch (Exception e) {
                              throw new Exception ('(2) DAO: Timestampumwandlungsproblem (java.sql.Date): '+e.message+"\n")
                              return false
                              }
                           break
                        // Timestampe
                        // HINT: in Oracle: TIMESTAMP PROBLEM
                        //
                        case oracle.sql.TIMESTAMP :
                           try{
                              this.setProperty(field.key,fieldvalue?.timestampValue())
                              }
                           catch (Exception e) {
                              throw new Exception ('(3) DAO: Oracle Timestampumwandlungsproblem (oracle.sql.TIMESTAMP): '+e.message+"\n")
                              return false
                              }
                           break
                        // weder TIMESTAMP oder DATE
                        //
                        default :
                           try{
                              this.setProperty(field.key,new java.sql.Timestamp(fieldvalue.getDate()))
                              }
                           catch (Exception e) {
                              throw new Exception ("(4) DAO: Feld:${fieldname}  Timestampumwandlungsproblem (unbekannter Type): "+e.message+"\n")
                              return false
                              }
                        }
                     break
                  // Sonstige Datenfelder
                  //
                  default :
                     this.setProperty(field.key,fieldvalue)
                  }
               writeLog("get: Feld: '${field.key}' -> '${fieldvalue}' type: ${field.value?.type}")
               }
            catch (Exception e){
               throw new Exception("DAO: DB->domain Feld:${fieldname} "+e.message+"\n")
               }
            }
         else {
            setProperty(field.key,null)
            }
         }

      }



  /**
   * Liest genau einen Datenbankrecord ueber einen Schluesselwert
   * in die Domain.
   *
   *
   * @param: id     inhaellt den Eideutigen Schluesse.
   *                HINT: Es liegt in der Verantwortung des Programmierers
   *                das dieser Tatschlich eindeutig ist. Es wird das erste Vorkommen
   *                geliefert.
   * 
   * @param: over   ID wird mit dem angegebenen Feld assoziiert.
   *                Darf nur mit option 'id' angewendet werden.
   * 
   * @param: where  get wird mit der angegebenen where Klausel 
   *                durchgefhrt. Schliesst 'id' und 'over' auss.
   * 
   * @return: true:  Datensatz wurde gefunden (Attribute des BO werden befllt
   *          false: Datensatz wurde nicht gefunden (Attribute des BO sind NULL)
   *
   * HINT:
   *    a) Diese Methode arbeite mit den Metadaten (meta) der Domain zusammen.
   *       Der Feldname wird ueber das Property dbfield ermittelt. Ist dieses nicht
   *       angegeben, wird der keywert verwendet.
   *    b) In der Domain kann eine Closure Namens after_read angegeben werden
   *       Dieses wird nachdem die Felder befllt worden sind aufgerufen und
   *       kann zur Veraenderung benutz werden.
   *
   */
   Boolean get(param) {

      // Kein Parametger ist ganz schlecht...
      assert (param != null) : "Methode get bentigt parameter "+this.getTablename()
      this.writeMode = MODE_UNKNOWN

      String listClause = giveFieldList()

      def id
      String keyField
      String where
      def rec 

      if (param instanceof Map) { // es wurde eine Map bergeben
         if (param.where != null) {
            keyField = param.get('where',' (1=1) ')
            this.lastSql = "select ${listClause} from ${tablename} WHERE ${keyField}"
            rec = this.db.firstRow(lastSql)
            }
         else {
            id = param.get('id',null)
            if (param.over!= null) {
               keyField = param.over == null ? this.meta?.idfield['db'] : param.over
               }
            else {
               keyField = this.meta?.idfield['db']
               }
            this.lastSql = "select ${listClause} from ${tablename} WHERE ${keyField} = ?"
            rec = this.db.firstRow(lastSql,[id])
            }
         }
      else { // es wurde nur ein wert bergeben
         id = param
         keyField = this.meta?.idfield['db']
         this.lastSql = "select ${listClause} from ${tablename} WHERE ${keyField} = ?"
         rec = this.db.firstRow(lastSql,[id])
         }

      writeLog("get: Keyfield: ${keyField}")
      writeLog("get: ${lastSql} [${id}]")

      // abfangen wenn nicht gefunden
      // Alle Felder des BO weden auf NULL gesetzt
      //
      if (rec == null) {
         this.meta?.fields.each {field->
            String fieldname = field.value?.dbfield == null ? field.key : field.value?.dbfield
            this.setProperty(field.key,null)
            }

          return false
          }

      // befuellen buisiness Object
      //

      fillDomainFields(rec)

      this.writeMode = MODE_READ

      // Eventaufruf:
      // Nach dem lesen...
      //
      this.meta?.after_read?.call()
      return true
      }

  /**
   * 
   *  Liefert die Whereclause mit dem Idfield fuer ein prepared statment
   * 
   *  usage mySql += getWhere
   * 
   *  @return: Whereclause z.B.: "WHERE myID = ?"
   * 
   */ 
   def getWhere = {
      String where = ' WHERE '
      where += this.meta?.idfield['db']+' = ? '
      }

  /**
   * 
   *  Liefert den Wert des Id-Feldes vom BO
   * 
   *  usage:   myID = employees.getIdValue()
   * 
   * 
   *  @return: Wert des ID-Feldes
   */ 
   def getIdValue = {

      def value = null

      this.meta?.fields.each {field->
         if (field.value.dbfield == this.meta?.idfield['db']) {
            value = this.getProperty(field.key)
            }
         }
      return value
      }

  /**
   * 
   *  + Setzt alle persistenten Felder der POGO auf null
   *  + Loescht Fehlermeldungsliste
   *
   *  usage: employees.clear()
   * 
   */ 
   def clear = {
      this.meta?.fields.each {field->
         String fieldname = field.value?.dbfield == null ? field.key : field.value?.dbfield
         this.setProperty(field.key,null)
         }
      clearErrors()
      }

  /**
   * 
   *  Lschen eines Datensatzes aus der Datenbank
   *  Die ID des BO wird aus dem BO ermittelt
   * 
   *  usage: employees.delete()
   * 
   *  Beispiel:
   *     employee.EMPLOYEE_ID = 100
   *     employee.delete()
   * 
   */ 
   def delete = {

      String deleteStm = "delete from ${this.tablename} "
      List deleteValues = []
      Integer cnt

      deleteStm += this.getWhere()
      deleteValues.add(this.getIdValue())
      this.lastSql = deleteStm+" ["+deleteValues.join(', ')+']'

      try {
         this.meta?.before_write?.call(mode:MODE_DELETE)
         }
      catch (Exception e) {
         throw new Exception ("${this.getClass().getName()}:: Es trat ein Fehler im  before_write von Event auf\n\"${e.message}\"")
         }
      cnt = db.executeUpdate(deleteStm,deleteValues)
      writeMode = MODE_DELETE
      writeLog("Loeschen : ${writeMode}\n${lastSql}")

      return (cnt == 0 ? false : true)
      }

  /**
   * 
   *  lscht alle Datensaetze aus der gegebenen Tabelle
   * 
   *  @param: where     Whereclause
   * 
   *  usage: cntRecords = deleteAll(where:'LOVCLASS = \'GEBUEHRENSTUFE\'')
   * 
   */ 
   def deleteAll = {param->

      assert param.where != null : "deleteAll muss Option where  haben "
      String stm = "delete from ${this.tablename} where "+param.where
      return db.executeUpdate(stm)
      }

   /**
    * 
    *  Persistieren des Objekts
    * 
    *  Das POGO wird in die Datenbank geschrieben
    *  Es wird ein update basierend auf dem Primary Key (idfield)
    *  versucht, konnte kein Update geschrieben werden wird ein insert geschrieben
    * 
    *  @param:  "mode"           ::= [insert|update]?
    *  @param:  "include"        ::= [fieldname,...]*
    * 
    *  @return: true: Die DB Operation war erfolgreich, sonst false
    * 
    *  HINT: Vor dem eigentlichen Schreibvorgang wird, wenn in der Domain definiert
    *        der Closure before_write aufgerufen. Dieser kann die Attribute der Domain
    *        Veraendern.
    *  HINT: Parameter include
    *        beinhaltet Feldnamen welche bearbeitet werden sollen.
    *        Wird dieser Parameter nicht bergeben, so werden alle Felder der 
    *        Metabeschreibung verwendet.
    * 
    */
   Boolean save (Map param) {
      if (param == null) param = [:]

      // wurde in der Domain ein Closure namens validate deklariert
      // so wird dieses ausgefhrt. Liefert dieses eine Fehlemeldung
      // wird das Programm abgebrochen und sqlError gesetzt.
      //     
      if (this.meta?.validate != null) {
         try {   // wir versuchen validate aufzurufen
            sqlError = this.meta?.validate.call(mode:param.get('mode',''))
            this.lastSql=''
            if ((sqlError?:'') != '') {
               return false
               }
            }
         catch (Exception e){
            throw new Exception ("DataAccessObject::validate (domain)\n${e.message}\n${sqlError}")
            }
         }

     /**
      * Durchlaufen aller Datenfelder und durchfuehren
      * des before_write event auf Feldebene
      */
      this.meta?.fields.each {field->
         if (field.value?.before_write != null) {
            this.setProperty(field.key,field.value.before_write.call(this.getProperty(field.key)))
            }
         }
         
      // Event: Vor dem Schreiben...
      //
      try {
         this.meta?.before_write?.call(mode:param.get('mode',''))
         }
      catch (Exception e) {
         throw new Exception ("${this.getClass().getName()}:: Es trat ein Fehler im  before_write von Event auf\n\"${e.message}\"")
         }

      List insertValues = []
      List updateValues = []

      String updateStm = "update ${this.tablename} set "
      String insertStm = "insert into ${this.tablename} ("

      // Statuszaehler fuer schreibende Operation
      Integer cnt

      // ID Inhalt
      //
      def fieldvalue

      // Zusammenstellung von insert und update statment
      //
      this.meta?.fields.each {field->

         String fieldname = field.value?.dbfield == null ? field.key : field.value?.dbfield
         
         // befuellen Datenfelds
         fieldvalue = this.getProperty(field.key)

         if (param.include != null) {
            // Wenn include gestzt werden
            // nur die includierte Felder zur Bildung des SQL Listclause verwendet
            // Das PK-Field wird auf jeden Fall eingebaut.
            if (field.key != (this.meta.idfield['domain']?:'')) {
               if (!param.include.grep(field.key)) return
               }
            }

         if (fieldvalue != null) {
            switch (field.value?.type) {
               case 'Date' :
                  // Dateum wird in SQL-Timestamp ungewandelt und auch Uhrzeit zu haben.
                  java.sql.Timestamp auxDate = new java.sql.Timestamp(fieldvalue.getTime())
                  fieldvalue = auxDate
                  break
               case 'OraDate' :
                  // Dateum wird in SQL-Timestamp ungewandelt und auch Uhrzeit zu haben.
                  java.sql.Timestamp auxDate = new java.sql.Timestamp(fieldvalue.getTime())
                  fieldvalue = auxDate
                  break
               }
            }

         insertValues.add(fieldvalue)
         updateValues.add(fieldvalue)
         if (updateValues.size() > 1) {
            updateStm += ', '
            insertStm += ', '
            }

         updateStm += "${fieldname} = ?"
         insertStm += fieldname
         }


      // Endfertigung Update
      //
      updateStm += this.getWhere()
      updateValues.add(this.getIdValue())
      // Endfertien Insert
      //
      insertStm += ') values ('
      insertValues.size().times() {insertStm += (it==0 ? '?' : ',?') }
      insertStm += ')'
      switch (param.get('mode','')) {
         case '' :
            // Wir versuchen ein update
            this.lastSql = updateStm+" ["+updateValues.join(', ')+']'
            try {
               cnt = db.executeUpdate(updateStm,updateValues)
               }
            catch (Exception e) {
               this.sqlError = "smartinsert(update): ${e.message}\n${insertStm},${insertValues.join(',')}"
               return false
               }
            this.writeMode = MODE_UPDATE

            // Wenn Zaehler 0 wurde nicht geschrieben
            // insert
            if (cnt == 0) {
               this.lastSql = insertStm+" ["+insertValues.join(', ')+']'
               try {
                  cnt = db.executeUpdate(insertStm,insertValues)
                  }
               catch (Exception e) {
                  this.sqlError = "smartinsert(insert): ${e.message}\n${insertStm},${insertValues.join(',')}"
                  return false
                  }

               this.writeMode = MODE_INSERT
               }
            break

         case MODE_INSERT :
            this.lastSql = insertStm+" ["+insertValues.join(', ')+']'
            cnt=1
            try {
               cnt = db.executeUpdate(insertStm,insertValues)
               }
            catch (Exception e) {
               this.sqlError = "insert: ${e.message}\n${insertStm}\n[${insertValues.join(',')}]"
               cnt=0
               return false
               }
            this.writeMode = MODE_INSERT
            break

         case MODE_UPDATE :
            this.lastSql = updateStm+" ["+updateValues.join(', ')+']'
            try {
               cnt = db.executeUpdate(updateStm,updateValues)
               }
            catch (Exception e) {
               this.sqlError = "update: ${e.message}\n${updateStm}\n[${updateValues.join(',')}]"
               return false
               }
            this.writeMode = MODE_UPDATE
            break

         default :
            this.writeMode = MODE_UNKNOWN
         }
      writeLog("save: writeMode: ${writeMode}\n${lastSql}")

      // Status zurckgeben
      //
      if (cnt == 0) {
         this.writeMode = MODE_ERROR
         return false
         }
      else {
         // Event: Vor dem Schreiben...
         //
         try {
            this.meta?.after_write?.call(mode:param.get('mode',''))
            }
         catch (Exception e) {
            throw new Exception ("${this.getClass().getName()}:: Es trat ein Fehler im  before_write von Event auf\n\"${e.message}\"")
            }

         
         return true
         }
      }

  /**
   *  
   *  In den Metadaten einer Domain kann der parameter prompt
   *  deklariert werden. Diese Funktion gibt diesen zurueck.
   *  Sollte kein Prompt definiert sein, wird eine Leerstring zurueckgegeben.
   *  
   *  @param   name     Feldname
   *  @return  Prompt oder Leerstring
   */  
   def getPrompt = {name->
      def field
      try {
         field = this.meta.fields.get(name)
         }
      catch (Exception e) {
         throw new Exception ("Fehler beim Prompt $name")
         }
      return (field.prompt?:'')
      }

  /**
   * Ermitteln des DB Feldnamens einer Domain
   * auf basis des Domainfeldnamens
   *
   * @param name      Domainfeldname
   * 
   * @return Datenbankfeldname
   * 
   * Die Methode wirft einen Exception wenn der Name 
   * nicht existiert.
   * 
   */
   String getDbFieldName(String name) {
      def field 
      field = this.meta.fields.get(name)
      return (field.dbfield?:'')
      }


  /**
   * Ermitteln des Datentyps eines Domainfeldes
   * 
   * @param name     Doaminfeldname
   * 
   */
   String getType(String name) {
      def field 
      field = this.meta.fields.get(name)
      return (field.type?:'')
      }



  /** 
   * Einfuegen eines Values in das BO
   * 
   * Es wird ein Value als String uebergeben und
   * die Prozedure befuellt das BO
   *
   * @param value       Inhalt
   * @param fieldname   Name des Feldes 
   *
   * @return   Status wenn Fehler aufgetreten ist wird true sonst false zurueckgeliefert
   */
   Boolean setDomainField(value,String fieldname) {
      Boolean isOk = true

     /** 
      * Ermittle field aus Metadaten
      */
      def field = this.meta?.fields[fieldname]

      if (field == null) throw new Exception("${this.getClass().getName()}::setDomainField: Kann Feldname ${fieldname} in Metadaten (meta) nicht finden ")

     /**
      * Wenn der Uebergebene Wert aus dem CGI einen Inhalt hat
      * wird Typumwandlung durchgefhrt
      */
      if (value != null) {
         // Wurde Typ bestimmt dann Umwandel
         // sonst wird String  angenommen
         //
         if (field?.type != null) {
            switch (field?.type) {
               case 'Integer' :
                  try {
                     this.setProperty(fieldname,value.toInteger())
                     }
                  catch (Exception e){
                     sqlError = "Feld '${fieldname}' kann nicht in Zahl umgewandelt werden; Inhalt: '${value}'"
                     errors.add([field:fieldname,message:'Zahlenumwandlungsfehler bei Ganzzahlwert',value:value])
                     this.setProperty(fieldname,null)
                     isOk = false
                     }
                  break
               case 'Long' :
                  try {
                     this.setProperty(fieldname,value.toLong())
                     }
                  catch (Exception e){
                     sqlError = "Feld '${fieldname}' kann nicht in Zahl (Long) umgewandelt werden; Inhalt: '${value}'"
                     errors.add([field:fieldname,message:'Zahlenumwandlungsfehler bei Ganzzahlwert (Long)',value:value])
                     this.setProperty(fieldname,null)
                     isOk = false
                     }
                  break
               case 'Double' :
                  try {
                     String auxValue = value
                     if (DECIMALPOINT_IS_COMMA) auxValue = (auxValue=~/\,/).replaceFirst('.')
                     this.setProperty(fieldname,auxValue.toDouble())
                     }
                  catch (Exception e){
                     sqlError = "Feld '${fieldname}' kann nicht in Zahl umgewandelt werden; Inhalt: '${value}'"
                     errors.add([field:fieldname,message:'Zahlenumwandlungsfehler bei Dezimalzahl',value:value])
                     this.setProperty(fieldname,null)
                     isOk = false
                     }

                  break
               case 'Date' :
               case 'OraDate' :
                  Date theDate
                  DateConverter dc = new DateConverter()
                  try{
                     theDate = dc.fromString(value)
                     this.setProperty(fieldname,theDate)
                     }
                  catch (Exception e) {
                     theDate = null
                     sqlError = "Feld '${fieldname}' kann nicht in Datum umgewandelt werden; Inhalt: '${value}'"
                     errors.add([field:fieldname,message:'Umwandlungsfehler bei Datum',value:value])
                     this.setProperty(fieldname,null)
                     isOk = false
                     }
                  break
               case 'String' :
                  this.setProperty(fieldname,value)
                  break
               default :
                   throw new Exception("DAO:setDomainField: Unbekannter Datentyp ${field?.type}")
                   break
               }
            }
         else { // kein Datentyp angegeben es wird String angenommen
            this.setProperty(fieldname,value)
            }
         }
      // Feldprfung Prfmethode anwenden
      // Wenn bishierhin OK
      //
      if (field?.nullable?:true == false) {
         if ((value?:'') =~ /^\s*$/) {
            errors.add([field:fieldname,message:'Feld darf nicht leer sein',value:value?:''])
            this.setProperty(fieldname,null)
            isOk = false
            }
         }

      if (field?.maxsize != null) {
         if ((value?:'').size()>field.maxsize) {
            errors.add([field:fieldname,message:"Feld darf nicht mehr als ${field.value.maxsize} Zeichen enthalten",value:''])
            this.setProperty(fieldname,null)
            isOk = false
            }
         }

      if (field?.validate != null) {
         String errorMessage = field?.validate?.call(value)
         if (errorMessage != '') {
            errors.add([field:fieldname,message:errorMessage,value:''])
            this.setProperty(fieldname,null)
            isOk = false
            }
         }
      if (!isOk) this.isOk = false
      return isOk
      }

  /** 
   * 
   *  Befuellt ein POGO aus dem CGI
   * 
   *  Die Routine geht davon aus, dass im URI die
   *  Parameternamen gleich dem Attributnamen im Objekt sind.
   * 
   *  @param: session      Sessionobjekt
   * 
   *  usage: lov.setFromCgi(pogo:lov)
   * 
   *  Die Routine wertet die Metadaten der Domain (meta) aus.
   *  Das Clousure validate wird durchgefhrt und ein Fehler
   *  gemeldet, wenn false zurckggeben wird.
   * 
   *  HINT:
   *     um eine korrekte Befllung zu gewhrleisten
   *     muss die Methode den Groovy Datentyp des Attributes wissen.
   *     Der Datentyp wird in der Meta Option (meta.fields) als Parameter
   *     type bezeichnet. Die Datentypendeklaration wird mit java.lang.integer usw.
   *     durchgefhrt.
   *     Wird keine Typendeklaration gefunden, wird String angenomen
   * 
   *     In der Domian knnen spezielle Verarbeitungen fuer CGI vereinbart werden.
   *     Derzeitig untertsttzt:
   *        * checkbox
   *     
   *  Untersteutze Datentypen:
   *    +  String
   *    +  Integer
   *    +  Long
   *    +  Double
   *    +  Date
   *    +  OraDate
   * 
   */ 
   def setFromCgi = {param->

      Boolean isOk = true
      errors = []
      String value 

      this.meta?.fields.each {field->
         if (param.include != null) {
            if (!param.include.grep(field.key)) return
            }
         value = param.request.getParameter(field.key)
         value = java.net.URLDecoder.decode(value?:'',formAceptCharset)

        /** 
         * In der Datenstruktur meta der Domain knnen Verarbeitungen
         * fuer die bernahme in CGI vereinbart werden
         * 
         * HINT:
         *    Derzeitig ist nur checkbox definiert
         */ 
         if (field.value?.cgi != null) {
            switch (field.value.cgi['type']) {
               case 'checkbox' : 
                  if ((value?:'') == '') {
                     value = field.value.cgi['default']
                     }
                  else {
                     value = field.value.cgi['checked']
                     }
                  break
               default :
                  throw new Exception ("DataAccessObject:: bei Feld '${field.key}' ungueltige cgi option '${field.value.cgi['type']}'")
                  break
               }
            }
        /** 
         * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
         *  set ist deprecated! und wird durch after_cgi ersetzt 
         * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
         */
         if (field.value.set != null) value = field.value.set.call(value)

        /**
         * Nach dem lesen aus de CGI wird 
         * wenn definiert die Closure aufgerufen
         * und das Ergebnis in das Value feld zurckgeschrieben
         */
         if (field.value.after_cgi != null) {
            value = field.value.after_cgi.call(value)
            }

        /** 
         * Wenn value ein Leerstring ("") enthaelt, wird
         * beim entsprechenden setting der Steuervariable EMPTY_IS_NULL
         * der Inhalt des Datenbankfeldes auf NULL gesetzt
         */
         if (EMPTY_IS_NULL && (value =~ /^\s*$/)) {
            value = null
            this.setProperty(field.key,null)
            }
          
        /** 
         * Befuellend des BO
         */
         try {
            isOk = setDomainField(value,field.key)
            }
         catch (Exception ex) {
            throw new Exception("${this.getClass().getName()}::setDomainField: ${ex.message}")
            }
         }
      return isOk
      }

  /**
   * 
   *  liefert den Vorgnger basierende auf einem aktuellen Datensatz
   *  HINT:
   *     Es wird davon ausgegangen, dass das Keyfeld Eindeutig ist
   * 
   *  @param:  key            Schluesselfeld in dem gesucht werden soll
   *  @param:  value          Aktueller Schluessel
   * 
   */
   def findPrivious = {param->

      if (param == null) param = [:]
      String tablename  = this.meta?.tablename
      String idfield    = this.meta?.idfield['db']

      String key        = param.get('key',idfield)
      String value      = param.get('value',this.getProperty(this.meta?.idfield['domain']))

      String sql = """
         select *
           from ${tablename}
          where ${idfield} = (select ${idfield}
                                  from ${tablename}
                                 where ${key} = (select MAX(${key})
                                                      from ${tablename}
                                                     where ${key} < '${value}'));
         """
      def rec = this.db.firstRow(sql)[idfield]
      this.get(rec)
      }

  /**
   * 
   *  liefert den Vorgnger basierende auf einem aktuellen Datensatz
   *  HINT:
   *     Es wird davon ausgegangen, dass das Keyfeld Eindeutig ist
   * 
   *  @param:  key            Schluesselfeld in dem gesucht werden soll
   *  @param:  value          Aktueller Schluessel
   * 
   */
   def findNext = {param->
      if (param == null) param = [:]
      String tablename  = this.meta?.tablename
      String idfield    = this.meta?.idfield['db']

      String key        = param.get('key',idfield)
      String value      = param.get('value',this.getProperty(this.meta?.idfield['domain']))

      String sql = """
         select *
           from ${tablename}
          where ${idfield} = (select ${idfield}
                                  from ${tablename}
                                 where ${key} = (select MIN(${key})
                                                      from ${tablename}
                                                     where ${key} > '${value}'));
         """

      def rec = this.db.firstRow(sql)[idfield]
      this.get(rec)
      }

  /**
   * 
   *  liefert den Primary-Key des ersten Datensatzes
   *  HINT:
   *     Es wird davon ausgegangen, dass das Keyfeld Eindeutig ist
   * 
   *  @param:  key            Schluesselfeld in dem gesucht werden soll
   * 
   */
   def findFirst = {param->
      if (param == null) param = [:]
      String tablename  = this.meta?.tablename
      String idfield    = this.meta?.idfield['db']

      String key        = param.get('key',idfield)
      
      String sql = """
         select *
           from ${tablename}
          where ${idfield} = (select ${idfield}
                                  from ${tablename}
                                 where ${idfield} = (select MIN(${key})
                                                      from ${tablename}));
         """
      def rec = this.db.firstRow(sql)[idfield]
      this.get(rec)
      }

  /**
   * 
   *  liefert den Primary-Key des letzten Datensatzes
   *  HINT:
   *     Es wird davon ausgegangen, dass das Keyfeld Eindeutig ist
   * 
   *  @param:  key            Schluesselfeld in dem gesucht werden soll
   *  @param:  value          Aktueller Schluessel
   * 
   */
   def findLast = {param->
      if (param == null) param = [:]

      String tablename  = this.meta?.tablename
      String idfield    = this.meta?.idfield['db']

      String key        = param.get('key',idfield)

      String sql = """
         select *
           from ${tablename}
          where ${idfield} = (select ${idfield}
                                  from ${tablename}
                                 where ${key} = (select MAX(${key})
                                                      from ${tablename}));
         """
      def rec = this.db.firstRow(sql)[idfield]
      this.get(rec)
      }

  /**
   * Durchlaufen aller Datensaete der 
   * uebergabe auf Recordbasis.
   * 
   */
   void findAll (Closure closure) {
      String sql = 'select * from '+this.meta?.tablename
      db.eachRow(sql) {
         closure(it)
         }
      }
  
  /**
   * 
   *  Basierend auf einer where Klausel
   *  werden die Datensaetze geholt und in
   *  das DAO geschrieben.
   * 
   *  @param:  where          Wherestatment
   *                          HINT: wohen das Schluesselwort 'where'
   *  @param:  orderby        Order by Klausel
   *  @param:  closure
   * 
   *  usage:
   *  Employees emp = new Employees(db:dbh)
   *  emp.eachDomain('EMPLOYEE_ID > 200') {domain->
   *     println domain.firstName+' '+domain.lastName
   *     }
   */
   void eachDomain(String where,String orderby,Integer max,Integer offset,Closure closure) {

      // Sql Statment zusammenbauen
      String whereclause = ''
      String orderbyclause = ''
      Long cntRec = max?:0L
      Long cntOff = offset?:0L

      if (where != null)      whereclause    = "where ${where}"
      if (orderby != null)    orderbyclause  = "order by ${orderby}"

      String sql = "select * from ${this.meta?.tablename} ${whereclause} ${orderbyclause} "

     /**
      * starten Datenstrom
      */
      db.eachRow(sql) {row->

        /**
         * Ueberlesen wenn Offset gewnscht
         */
         if (cntOff > 0) {
            cntOff--
            return
            }
         
        /**
         * Wenn maximale Anzal gewuenscht
         * wird Recordzaehler minimiert
         * Wenn nicht Zhler minimieren
         */
         if(max == null) 
            cntRec = 1
         else 
            cntRec--
         
         if (cntRec >= 0) {
            fillDomainFields(row)
            this.meta?.after_read?.call()
            // Aufrufen Closure
            closure(this)
            }
         }
      }

   void eachDomain(Closure closure) {eachDomain(null,null,null,null,closure)}
   void eachDomain(String where,Closure closure) {eachDomain(where,null,null,null,closure)}
   void eachDomain(String where,String orderby, Closure closure) {eachDomain(where,orderby,null,null,closure)}

  /**
   * EachDomain mit benannten Parameter
   * 
   * @param "where"        Where Statment
   * @param "orderby"      Felder mit Sortierkriterien (optional)
   * @param "conditions"   Key Values der Kriterien (optional)
   *                       [jobID:'PU_ACCOUNT']
   * @param "offset"       Anzahl der zu berlesenden Datenstze (optional)
   *                       Vorgabewert 0
   * @param "max"          Maximale Anzal der Datenstze (optional)
   *                       Vorgabewert unbeschrnkt
   * 
   */
   void eachDomain(Map param,Closure closure) {
      if (param == null) param = [:]
      String  where     = param.get('where')
      String  orderby   = param.get('orderby')
      Integer offset    = param.get('offset',0)
      Integer max       = param.get('max',null)
      def criteria  = param.get('criteria',null)
      // Wurde Kriterien uebergeben
      if (criteria) {
         if (criteria instanceof QueryBuilder) {
            where = criteria.getWhere()
            this.lastSql = where
            }
         else {
            QueryBuilder query = new QueryBuilder(domain:this)
            query criteria
            where = query.getWhere()
            this.lastSql = where
            }
         }
     
      eachDomain(where,orderby,max,offset,closure)
      }
   

  /**
   * Durchlaufen einer Domain basierend auf
   * einer SQL Anweisung
   *  
   * @param:  sql
   * @param:  closure
   * 
   * usage:
   *   eachDomainWithSql('select * from employee') {
   *      ...
   *      }
   *   
   */
   void eachDomainWithSql(String sql,Closure closure) {

      // Datenstrom durchlaufen
      //
      db.eachRow(sql) {row->

         // befuellen buisiness Object
         //
         fillDomainFields(row)
         this.meta?.after_read?.call()
         // Aufrufen Closure
         closure(this)
         }
      }

  /**
   * Each Domain mit LIMIT
   * 
   * @param where          Wherestatment
   * @param orderby        Sortieranweisung
   * @param top            1. Datensatz
   * @param count          n Datensaetze
   * @param closure        Closure
   *
   */
   void eachDomainWithLimit(String where, String orderby, Integer top, Integer number, Closure closure) {
      // Sql Statment zusammenbauen
      String whereclause = ''
      String orderbyclause = ''
      Integer cnt = 0

      if (where != null)      whereclause    = "where ${where}"
      if (orderby != null)    orderbyclause  = "order by ${orderby}"

      String sql = "select * from ${this.meta?.tablename} ${whereclause} ${orderbyclause} "
      // Datenstrom durchlaufen
      //
      db.eachRow(sql) {row->
         cnt++ // Datensatzzaehlr erhhen
         if (cnt < top) return
         if (cnt > (top+number)) return

         // befuellen buisiness Object
         //
         fillDomainFields(row)
         this.meta?.after_read?.call()
         // Aufrufen Closure
         closure(this)
         }
      }

  /**
   * Liefert das ergebnis von Count
   */
   Integer sqlCount(Map param) {
      
      Integer retval = 0
      String sql = "select count(*) from ${this.meta.tablename}"
      if (param == null)  param=[:]
      
      if (param.get('where') != null) sql += " where ${param.where}"

      try {
         retval = this.db.firstRow(sql)[0]
         }
      catch(Exception ex) {
         throw new Exception("In ${this.getClass().getName()} bei sqlCount trats ein Fehler auf ${ex.message}")
         }

      return retval
      }

  /**
   * Liefert das Ergebnis von Sum als Double
   */
   Double sqlSum(Map param) {

      if (param == null) param = [:]

      if (param.get('field') == null) {
         throw new Exception("${this.getClass().getName()} fehlt der Parameter field")
         }

      String sql = "select SUM(${param.field}) from ${this.meta.tablename}"

      Integer retval = 0
      if (param.get('where') != null) sql += " where ${param.where}"

      try {
         retval = this.db.firstRow(sql)[0]
         }
      catch(Exception ex) {
         throw new Exception("In ${this.getClass().getName()} bei sqlCount trats ein Fehler auf ${ex.message}")
         }

      return retval
      }

  /**
   * Liefert das Ergebnis von Avg als Double
   */
   Double sqlAvg(Map param) {

      if (param == null) param = [:]

      if (param.get('field') == null) {
         throw new Exception("${this.getClass().getName()} fehlt der Parameter field")
         }

      String sql = "select AVG(${param.field}) from ${this.meta.tablename}"

      Integer retval = 0
      if (param.get('where') != null) sql += " where ${param.where}"

      try {
         retval = this.db.firstRow(sql)[0]
         }
      catch(Exception ex) {
         throw new Exception("In ${this.getClass().getName()} bei sqlCount trats ein Fehler auf ${ex.message}")
         }

      return retval
      }

  /**
   * Liefert das Ergebnis von Min als Double
   */
   Double sqlMin(Map param) {

      if (param == null) param = [:]

      if (param.get('field') == null) {
         throw new Exception("${this.getClass().getName()} fehlt der Parameter field")
         }

      String sql = "select MIN(${param.field}) from ${this.meta.tablename}"

      Integer retval = 0
      if (param.get('where') != null) sql += " where ${param.where}"

      try {
         retval = this.db.firstRow(sql)[0]
         }
      catch(Exception ex) {
         throw new Exception("In ${this.getClass().getName()} bei sqlCount trats ein Fehler auf ${ex.message}")
         }

      return retval
      }

  /**
   * Liefert das Ergebnis von Min als Double
   */
   Double sqlMax(Map param) {

      if (param == null) param = [:]

      if (param.get('field') == null) {
         throw new Exception("${this.getClass().getName()} fehlt der Parameter field")
         }

      String sql = "select MAX(${param.field}) from ${this.meta.tablename}"

      Integer retval = 0
      if (param.get('where') != null) sql += " where ${param.where}"

      try {
         retval = this.db.firstRow(sql)[0]
         }
      catch(Exception ex) {
         throw new Exception("In ${this.getClass().getName()} bei sqlCount trats ein Fehler auf ${ex.message}")
         }

      return retval
      }

  /**
   * liefert eine Closure der durch
   * alle meta Felder laeuft
   * 
   */
   void allFields(Closure closure) {
      this.meta.fields.each{field->
         closure.call(field)
         }
      }

   } // END OF OBJECT